// primitives.js
//   This code uses the libraries from the WebGL book,

//----------- declare globals  ------------------------------------------------------

var gl;              // GL rendering context
var gModelViewMatrix = new Matrix4();  // create new 4x4 matrix
var g_time = 0;      // global elapsed time, as measured in seconds
var canvas;          // canvas that WebGL will use to draw
var gCamera;
var gRotating = false;  // turntable rotation

var gLight = { direction: new Vector4() };
var gGroundWidth = 400;

var gCanvasWidth = 0;
var gCanvasHeight = 0;

// geometry
var gCube;
var gPlane;
var gMesh;

// shaders
var gMeshProg = new ShaderProgram();
var gColourProg = new ShaderProgram();

// textures
var gMeshTexture0 = null;
var gMeshTexture1 = null;
var gMeshTexture2 = null;

var gRotTheta = 0;

//----------- main()   --------------------------------------------------------------

function main() {    
    canvas = mySetup();       // setup shaders, send vertices to the GPU
    
    // Set clear color and enable hidden surface removal
    gl.enable(gl.DEPTH_TEST);
    gl.enable(gl.CULL_FACE);
    gl.cullFace(gl.BACK);

    // Register the event handler to be called on key press
    document.onkeydown = function(ev){ keydown(ev); };

    animate();
}

//----------- animate()   --------------------------------------------------------------

function animate() {
  var tick = function() {                // create the callback function
    update();
    draw();
    requestAnimationFrame(tick, canvas); // Request that the browser calls tick
  };
  tick();                // now call it once. Further calls will be via the requests
}

//----------- update()   --------------------------------------------------------------

var g_startTime = Date.now();   // last time the function was called, in milliseconds

function update() {
    if (gRotating) {
      dt = (Date.now() - g_startTime)/1000;
      g_startTime = Date.now();
      g_time += dt;
      gRotTheta = (g_time * 40)%360;
    }
}

//----------- draw()   --------------------------------------------------------------

function draw() {
    // draw actual scene
    gl.viewport(0, 0, gCanvasWidth, gCanvasHeight);
    
    // Clear color and depth buffer
    gl.clearColor(0.7, 0.80294117647, 0.82745098039, 1.0);
    gl.clear(gl.COLOR_BUFFER_BIT | gl.DEPTH_BUFFER_BIT);

    if (gMeshProg.loaded)
    {
        drawScene(gMeshProg, gCamera);
    }
}

function drawScene(prog, camera)
{
    prog.Bind();

    // setup projection matrix
    var proj_matrix = camera.getProjectionMatrix();
    gl.uniformMatrix4fv(prog.uniforms.u_ProjectionMatrix, false, proj_matrix.elements);

    // Set the eye point and the viewing volume
    gModelViewMatrix = camera.getViewMatrix();

      // draw ground
    gMeshTexture0.bind(gl.TEXTURE0);              // marble texture
    gl.uniform1i(prog.uniforms.u_AlbedoTex, 0);
    var ground_colour = [0.8, 0.6, 0.6, 1];
    gl.uniform4f(prog.uniforms.u_FragColor, ground_colour[0], ground_colour[1], ground_colour[2], ground_colour[3]);
    gl.uniform1f(prog.uniforms.u_DistortionTime, 0.0);
    gl.uniform1f(prog.uniforms.u_DistortionAmp, 0.0);
    gPlane.bindVertices(prog.attributes.a_Position);
    gPlane.bindTexCoords(prog.attributes.a_TexCoord);
    drawGround(prog);

      // draw bunny
    if (gMesh.loaded) {           // make sure the mesh is fully loaded before drawing
        gMeshTexture1.bind(gl.TEXTURE1);
        gl.uniform1i(prog.uniforms.u_AlbedoTex, 1);
        gMesh.bindVertices(prog.attributes.a_Position);
        gMesh.bindVertices(prog.attributes.a_Normal);
        gMesh.bindTexCoords(prog.attributes.a_TexCoord);
        var mesh_colour = [0.9, 0.9, 0.9, 1];
        gl.uniform4f(prog.uniforms.u_FragColor, mesh_colour[0], mesh_colour[1], mesh_colour[2], mesh_colour[3]);
	gl.uniform1f(prog.uniforms.u_DistortionTime, g_time);
        gl.uniform1f(prog.uniforms.u_DistortionAmp, 0.05);
        drawMesh(prog,50.0);
    }

     // draw cube
    gMeshTexture2.bind(gl.TEXTURE2);              // marble texture
    gl.uniform1i(prog.uniforms.u_AlbedoTex, 2);
    var cube_colour = [1.0, 1.0, 1.0, 1];
    gl.uniform4f(prog.uniforms.u_FragColor, cube_colour[0], cube_colour[1], cube_colour[2], cube_colour[3]);
    gl.uniform1f(prog.uniforms.u_DistortionTime, 0.0);
    gl.uniform1f(prog.uniforms.u_DistortionAmp, 0.0);
    gCube.bindVertices(prog.attributes.a_Position);
    gCube.bindTexCoords(prog.attributes.a_TexCoord);
    drawCube(prog);

}

function drawCube(prog)
{
    var cube_size = 2;
    pushMatrix(gModelViewMatrix);

    gModelViewMatrix.translate(7, 4, 0);
    gModelViewMatrix.scale(cube_size, cube_size, cube_size);
    gl.uniformMatrix4fv(prog.uniforms.u_ModelViewMatrix, false, gModelViewMatrix.elements); // send matrix over to gpu

    gCube.draw();

    gModelViewMatrix = popMatrix();
}

function drawGround(prog)
{
    var ground_w = gGroundWidth;
    
    pushMatrix(gModelViewMatrix);
    gModelViewMatrix.scale(ground_w, ground_w, ground_w);
    gModelViewMatrix.translate(-0.5, 0.0, -0.5);
    gl.uniformMatrix4fv(prog.uniforms.u_ModelViewMatrix, false, gModelViewMatrix.elements); // send matrix over to gpu
    gPlane.draw();
    gModelViewMatrix = popMatrix();
}

function drawMesh(prog,mesh_scale)
{
    pushMatrix(gModelViewMatrix);
    gModelViewMatrix.scale(mesh_scale, mesh_scale, mesh_scale);
    gModelViewMatrix.rotate(gRotTheta, 0, 1, 0);
    gl.uniformMatrix4fv(prog.uniforms.u_ModelViewMatrix, false, gModelViewMatrix.elements); // send matrix over to gpu
    gMesh.draw();
    gModelViewMatrix = popMatrix();
}

//----------- pushMatrix()   -----------------------------------------------------------

var g_matrixStack = []; // Array for storing a matrix
function pushMatrix(m) { // Store the specified matrix to the array
  var m2 = new Matrix4(m);
  g_matrixStack.push(m2);
}

//----------- popMatrix()   -----------------------------------------------------------

function popMatrix() { // Retrieve the matrix from the array
  return g_matrixStack.pop();
}

//----------- keydown()   -----------------------------------------------------------

function keydown(ev) {
    if(ev.keyCode == 65) { // 'a' key pressed
        gCamera.rotate(-10, 0, 1, 0);
    } else 
    if (ev.keyCode == 83) { // 's' key pressed
        gCamera.rotate(10, 0, 1, 0);
    } else
    if (ev.keyCode == 32) { // space key pressed
        gRotating = !gRotating;   // toggle rotation
        g_startTime = Date.now();   // last time the function was called, in milliseconds
    } else { return; }
}

//----------- mySetup()   -----------------------------------------------------------

function mySetup() {
    var canvas = document.getElementById('webgl');
    gl = getWebGLContext(canvas);          // get rendering context 
    if (!gl) throw 'Failed to get the rendering context for WebGL';

    gCanvasWidth = canvas.width;
    gCanvasHeight = canvas.height;

    setupCamera();
    setupLight();
    buildGeometry();
    setupShaders();
    loadTextures();

    return canvas;
}

function setupCamera()
{
    var fovy = 30;
    var aspect_ratio = gCanvasWidth / gCanvasHeight;
    var near = 1;
    var far = 100;
    
    var eyeX = 8;
    var eyeY = 20;
    var eyeZ = 30;

    var centerX = 0;
    var centerY = 5;
    var centerZ = 0;

    var upX = 0;
    var upY = 1;
    var upZ = 0;

    gCamera = new Camera();
    gCamera.buildPerspectiveCam(fovy, aspect_ratio, near, far,
                                 eyeX, eyeY, eyeZ,
                                 centerX, centerY, centerZ,
                                 upX, upY, upZ);
}

function setupLight()
{
    // a directional light source
    gLight.direction.elements[0] = 25;
    gLight.direction.elements[1] = 40;
    gLight.direction.elements[2] = 15;
    gLight.direction.elements[3] = 0;
}

function setupShaders()
{
    // mesh shader
    // samples from a texture map
    gMeshProg.Build(gl, 'shaders//mesh_vs.glsl', 'shaders//mesh_fs.glsl',
        function () {
            gMeshProg.SetupUniform('u_ModelViewMatrix');
            gMeshProg.SetupUniform('u_ProjectionMatrix');
            gMeshProg.SetupUniform('u_FragColor');
            gMeshProg.SetupUniform('u_AlbedoTex');
            gMeshProg.SetupUniform('u_DistortionTime');
            gMeshProg.SetupUniform('u_DistortionAmp');

            gMeshProg.SetupAttribute('a_Position');
            gMeshProg.SetupAttribute('a_TexCoord');
    });
}

function loadTextures()
{
    var tex_params = new TextureParams(gl);
    gMeshTexture0 = new Texture("textures//ground.png", tex_params);
    gMeshTexture1 = new Texture("textures//ubc2.png", tex_params);
    gMeshTexture2 = new Texture("textures//ubcTexture.png", tex_params);
}

// ---------------- build meshes ------------------------

function buildGeometry()
{
    buildCube();
    buildPlane();
    
    gMesh = new Mesh(gl);
    gMesh.loadMesh('meshes//bunny.obj');
//    gMesh.loadMesh('meshes//teapot.obj');
}

function buildCube()
{
    // setup cube
    var cube_verts = [
         // Front face
            -1.0, -1.0, 1.0,
             1.0, -1.0, 1.0,
             1.0, 1.0, 1.0,
            -1.0, 1.0, 1.0,

            // Back face
            -1.0, -1.0, -1.0,
            -1.0, 1.0, -1.0,
             1.0, 1.0, -1.0,
             1.0, -1.0, -1.0,

            // Top face
            -1.0, 1.0, -1.0,
            -1.0, 1.0, 1.0,
             1.0, 1.0, 1.0,
             1.0, 1.0, -1.0,

            // Bottom face
            -1.0, -1.0, -1.0,
             1.0, -1.0, -1.0,
             1.0, -1.0, 1.0,
            -1.0, -1.0, 1.0,

            // Right face
             1.0, -1.0, -1.0,
             1.0, 1.0, -1.0,
             1.0, 1.0, 1.0,
             1.0, -1.0, 1.0,

            // Left face
            -1.0, -1.0, -1.0,
            -1.0, -1.0, 1.0,
            -1.0, 1.0, 1.0,
            -1.0, 1.0, -1.0
    ];

    var cube_normals = [
            // Front face
             0.0, 0.0, 1.0,
             0.0, 0.0, 1.0,
             0.0, 0.0, 1.0,
             0.0, 0.0, 1.0,

             // Back face
             0.0, 0.0, -1.0,
             0.0, 0.0, -1.0,
             0.0, 0.0, -1.0,
             0.0, 0.0, -1.0,

             // Top face
             0.0, 1.0, 0.0,
             0.0, 1.0, 0.0,
             0.0, 1.0, 0.0,
             0.0, 1.0, 0.0,

             // Bottom face
             0.0, -1.0, 0.0,
             0.0, -1.0, 0.0,
             0.0, -1.0, 0.0,
             0.0, -1.0, 0.0,

             // Right face
             1.0, 0.0, 0.0,
             1.0, 0.0, 0.0,
             1.0, 0.0, 0.0,
             1.0, 0.0, 0.0,

             // Left face
             -1.0, 0.0, 0.0,
             -1.0, 0.0, 0.0,
             -1.0, 0.0, 0.0,
             -1.0, 0.0, 0.0,
    ];

    var cube_tex_coords= [
         // Front face
          0.0, 0.0,
          1.0, 0.0,
          1.0, 1.0,
          0.0, 1.0,

          // Back face
          1.0, 0.0,
          1.0, 1.0,
          0.0, 1.0,
          0.0, 0.0,

          // Top face
          0.0, 1.0,
          0.0, 0.0,
          1.0, 0.0,
          1.0, 1.0,

          // Bottom face
          1.0, 1.0,
          0.0, 1.0,
          0.0, 0.0,
          1.0, 0.0,

          // Right face
          1.0, 0.0,
          1.0, 1.0,
          0.0, 1.0,
          0.0, 0.0,

          // Left face
          0.0, 0.0,
          1.0, 0.0,
          1.0, 1.0,
          0.0, 1.0,
    ];

    var cube_indices = [       // indices 
        0, 1, 2,        0, 2, 3,    // Front face
        4, 5, 6,        4, 6, 7,    // Back face
        8, 9, 10,       8, 10, 11,  // Top face
        12, 13, 14,     12, 14, 15, // Bottom face
        16, 17, 18,     16, 18, 19, // Right face
        20, 21, 22,     20, 22, 23  // Left face
    ];

    gCube = new Mesh(gl);
    gCube.setVertices(cube_verts);
    gCube.setNormals(cube_normals);
    gCube.setTexCoords(cube_tex_coords);
    gCube.setIndices(cube_indices);
}

function buildPlane()
{
    // setup plane
    var plane_verts = [ 0, 0, 0,
                        1, 0, 0,
                        1, 0, 1,
                        0, 0, 1];

    var plane_normals = [0, 1, 0,
                         0, 1, 0,
                         0, 1, 0,
                         0, 1, 0];

    var plane_tex_coords = [0, 0,
                     2, 0,
                     2,2,
                     0, 2];
    var plane_indices = [0, 2, 1,
                         2, 0, 3];

    gPlane = new Mesh(gl);
    gPlane.setVertices(plane_verts);
    gPlane.setNormals(plane_normals);
    gPlane.setTexCoords(plane_tex_coords);
    gPlane.setIndices(plane_indices);
}
